PROCEDURE CrossTab   && UDS for Capturing Parameters for CrossTab
*-----------------------------------------------------------------------
* Capture CrossTab parameters for COM_XTAB
*
* Synopsis:
* 
* CROSSTAB( )
* 
* Description:
* 
* The CrossTab() function displays the front end to capture the 
* parameters for the COM_XTAB (compute crosstab) procedure.
* The CrossTab() function is called from within the QBE design
* surface by placing the function CROSSTAB() in any one of the field
* skeletons.  CrossTab() will not work within the file pothandle.
* 
* The user then fills in the optional title, the column to summarize
* across the top, the column to summarize along the side, the column
* to compute in the middle, and the type of operation (SUM,CNT,MAX,MIN).
* 
* If the user presses F2 to perform the crosstab, the CrossTab() will
* remove the CROSSTAB() reference from the skeleton and keyboard in
* the commands to transfer to the Applications Panel in the Control
* Center and run EXECCROS.
* 
* If the user presses Ctrl-End to save, but not execute, the CrossTab(), 
* the CrossTab() function will save the parameters to the CROSSTAB.MEM 
* file and clear out the CROSSTAB() reference from the skeletion.
* 
* If the user presses Escape to cancel, the CrossTab() will not save
* any of the values and clear out the CROSSTAB() reference from 
* the skeletion.
* 
* The summary field must be numeric.  The top and side fields can be
* any type but memo.
* 
* See also:
* 
* EXECCROS.PRG    Run the CrossTab and display the results
* COM_XTAB.PRG    Perform the actual CrossTab computations
* 
*-----------------------------------------------------------------------

  PRIVATE lc_talk, lc_cursor, lc_display, lc_status, lc_carry, lc_proc,;
  ln_typeahd

  SAVE SCREEN TO qbe
  IF SET("TALK") = "ON"
    SET TALK OFF
    lc_talk = "ON"
  ELSE
    lc_talk = "OFF"
  ENDIF
  lc_cursor = SET("CURSOR")
  SET CURSOR ON

  lc_escape = SET("ESCAPE")
  SET ESCAPE OFF
  lc_status = SET("STATUS")
  lc_safety = SET("SAFETY")
  SET SAFETY OFF

  IF TYPE("kn_esc") = "U"
    DO key_vars
  ENDIF

  ll_iscolor = IIF( SET( "COLOR" ) = "ON", .t., .f. )
  IF TYPE("SAYCOLOR") = "U"
    PUBLIC saycolor
    IF ll_iscolor
      saycolor = "rg+"
    ELSE
      saycolor = "w/n"
    ENDIF
  ENDIF
  IF TYPE("FILLCOLOR") = "U"
    PUBLIC fillcolor
    IF ll_iscolor
      fillcolor = "gb+"
    ELSE
      fillcolor = "n"
    ENDIF
  ENDIF
  IF TYPE("MESSCOLOR") = "U"
    PUBLIC messcolor
    messcolor = COLOR("MESSAGE")
    ln_slash = AT("/", messcolor)
    ln_plus  = AT("+", messcolor)
    IF ln_plus = 0
      messcolor = "+" + messcolor
    ELSE
      IF ln_plus > ln_slash
        messcolor = "+" + messcolor
      ENDIF
    ENDIF
  ENDIF

  DEFINE WINDOW Crosstab FROM 5,7 TO 18,73 DOUBLE
  DO shadowg WITH 5,7,18,73
  ACTIVATE WINDOW Crosstab

  DO XT_DisSays

  DO XT_IniVars

*-- Check to see if a numeric field is present in the DBF
  ll_numpres = .F.
  ln_fldcnt = 1
  DO WHILE .NOT. ISBLANK( FIELD( ln_fldcnt ) )
    IF TYPE( FIELD( ln_fldcnt ) ) $ "NF"
      ll_numpres = .T.
      EXIT
    ENDIF
    ln_fldcnt = ln_fldcnt + 1
  ENDDO
  IF ll_numpres                        && Numeric field found
    gc_actlist = "@M SUM,CNT,MAX,MIN"  && Give the full set of options
  ELSE
    gc_actlist = "@M CNT,MAX,MIN"      && Else allow all but SUM
  ENDIF

  ON KEY LABEL f2       DO XT_F2Exit
  ON KEY LABEL shift-f1 DO XT_FldPick  WITH VARREAD()
  ll_escape = .F.
  ll_edit = .T.
  ll_f2 = .F.
  DO WHILE ll_edit
    DO XT_DisGets
    READ
    ln_readkey = READKEY()
    DO CASE
      CASE ln_readkey = rn_Esc
        ll_escape = .T.
        ll_edit = .F.
      CASE ln_readkey = rn_CtrlEnd
        IF ISBLANK( gc_xtop ) .OR. ISBLANK( gc_xside ) .OR. ISBLANK( gc_xsum )
          ll_edit = .T.
          ACTIVATE SCREEN
          @ 24,0
          DO Beep
          lc_msg = "The top, side, and summary fields must be specified" + ;
                   " (Press SPACE)"
          @ 24, Center( lc_msg,80 ) SAY lc_msg COLOR &messcolor
          ln_tmp = INKEY(0)
          @ 24,0
          ACTIVATE WINDOW CrossTab
        ELSE
          ll_edit = .F.
        ENDIF
    ENDCASE
  ENDDO
  ON KEY LABEL f2
  ON KEY LABEL shift-f1

  DO XT_TrmVars                  && Trim the vars from the form
  lc_fullp = SET("FULLPATH")
  SET FULLPATH ON

  gl_XIsCC = XT_MakExt()        && Make EXECCROS.PRG is needed
  IF ll_f2
    gc_xdbf = DBF()             && Capture open DBF file
    DO XT_TrmVars                && Trim the vars from the form
    SAVE TO crosstab ALL LIKE gc_x*
    SET INSTRUCT OFF            && Prevent extra message when running app
    SET SAFETY OFF

    IF FILE( "crosstab.qbe" )   && Delete the existing CROSSTAB.QBE file
      ERASE crosstab.qbe
    ENDIF

    KEYBOARD "{Alt-L}S"                 && Layout menu, Save bar
    KEYBOARD "{Home}{Ctrl-Y}"           && Clear existing QBE name
    KEYBOARD "Crosstab{Ctrl-M}"         && Save as Crosstab.qbe
    KEYBOARD "{Ctrl-End}"               && Exit QBE
    KEYBOARD "{27}Y"                    && Exit the control center
    KEYBOARD "DO ExecCros{Ctrl-M}"      && Run ExecCros.PRG

  ELSE
    IF .NOT. ll_escape
      ll_edit = .F.
      gc_xdbf = DBF()
      DO XT_TrmVars              && Trim the vars from the form
      SAVE TO crosstab ALL LIKE gc_x*
    ENDIF
  ENDIF

  DEACTIVATE WINDOW Crosstab
  RELEASE WINDOW Crosstab
  RELEASE POPUP crosstab

  RELEASE gc_xtitl,gc_xtop,gc_xside,gc_xsum,gc_xoper
  RELEASE aa_lookup, params

  SET STATUS &lc_status.
  SET CURSOR &lc_cursor.
  SET ESCAPE &lc_escape.
  SET SAFETY &lc_safety.
  SET TALK &lc_talk.
  SET FULLPATH &lc_fullp

  RESTORE SCREEN FROM qbe
  RELEASE SCREEN qbe
RETURN
*-- EOF: CrossTab()


PROCEDURE XT_F2Exit  && UDS - CrossTab()
*-----------------------------------------------------------------------
* Handle the ON KEY LABEL f2 by setting the state variable LL_F2 to
* true and keyboarding a Ctrl-End into the CrossTab form.
*-----------------------------------------------------------------------
  ll_f2 = .T.
  KEYBOARD CHR( kn_CtrlEnd )
RETURN
*-- EOP: XT_F2Exit


PROCEDURE XT_DisSays && UDS - CrossTab()
*-----------------------------------------------------------------------
* Display the form with the appropiate colors.  Uses the PUBLIC
* variable FILLCOLOR to display in inside fill color.  Uses the PUBLIC
* variable SAYCOLOR to display the text information and lines.
*-----------------------------------------------------------------------
  @ 0,0 SAY REPLICATE( CHR(219), 65)
  @ 1,0 SAY REPLICATE( CHR(219), 65)
  @ 2,0 SAY REPLICATE( CHR(219), 65)
  @ 3,0 SAY REPLICATE( CHR(219), 65)
  @ 4,0 SAY REPLICATE( CHR(219), 65)
  @ 5,0 SAY REPLICATE( CHR(219), 65)
  @ 6,0 SAY REPLICATE( CHR(219), 65)
  @ 7,0 SAY REPLICATE( CHR(219), 65)
  @ 8,0 SAY REPLICATE( CHR(219), 65)
  @ 9,0 SAY REPLICATE( CHR(219), 65)
  @ 10,0 SAY REPLICATE( CHR(219), 65)
  @ 11,0 SAY REPLICATE( CHR(219), 65)
  @ 0,0 FILL TO 11,64 COLOR &fillcolor

*-- @ SAY Processing. -----------------------------------------------------
  @ 0,24 SAY "Cross Tabulation" COLOR &saycolor
  @ 1,0 SAY "" COLOR &saycolor
  @ 2,10 SAY "Title:" COLOR &saycolor
  @ 4,18 SAY "Top Field" COLOR &saycolor
  @ 4,28 SAY ">" COLOR &saycolor
  @ 5,11 SAY "" COLOR &saycolor
  @ 6,0 SAY "Side Field:" COLOR &saycolor
  @ 7,5 SAY "" COLOR &saycolor
  @ 7,11 SAY "" COLOR &saycolor
  @ 7,34 SAY "Summary Field:" COLOR &saycolor
  @ 8,5 SAY "V" COLOR &saycolor
  @ 8,11 SAY "" COLOR &saycolor
  @ 9,11 SAY "" COLOR &saycolor
  @ 9,34 SAY "  Operation:" COLOR &saycolor
  @ 10,11 SAY "" COLOR &saycolor
  @ 11,11 SAY "" COLOR &saycolor

RETURN               
*-- EOP: XT_DisSays


PROCEDURE XT_DisGets && UDS - CrossTab()
*--------------------------------------------------------------------
* Display the GET fields for the form.  Display the Navigation line
* before starting the GET processing.  Allow Shift-F1 picklists
* for the field selections.
*--------------------------------------------------------------------
  PRIVATE lc_errmsg
  lc_errmsg = "Try again"

*-- Perform the GETs
  @  2, 17 GET gc_xtitl  PICTURE "@S40";
    MESSAGE "Specify an optional heading for the CrossTab display";
    WHEN XT_NavDsp( 1 )

  @  4, 34 GET gc_xtop   PICTURE "@!S10" ;
    MESSAGE "Specify the field to summarize along the top of the CrossTab";
    WHEN XT_NavDsp( 2 );
    VALID REQUIRED IsInView( gc_xtop ) .AND. TYPE( gc_xtop )<>"M";
    ERROR IIF( TYPE(gc_xtop)="M",;
                  "Memo fields are not allowed with CrossTab",;
                  "The field specified is not in the current view")


  @  9,  0 GET gc_xside  PICTURE "@!S10" ;
    MESSAGE "Specify the field to summarize along the side of the CrossTab";
    WHEN XT_NavDsp( 3 );
    VALID REQUIRED IsInView( gc_xside ) .AND. TYPE( gc_xside )<>"M";
    ERROR IIF( TYPE(gc_xside)="M",;
                  "Memo fields are not allowed with CrossTab",;
                  "The field specified is not in the current view")

  @  8, 34 GET gc_xsum   PICTURE "@!S10" ;
    MESSAGE "Specify the field to operate on in the body of the CrossTab";
    WHEN XT_NavDsp( 4 );
    VALID REQUIRED IsInView( gc_xsum ) .AND. TYPE( gc_xsum )<>"M";
    ERROR IIF( TYPE(gc_xsum)="M",;
                  "Memo fields are not allowed with CrossTab",;
                  "The field specified is not in the current view")

  @ 10, 39 GET gc_xoper  PICTURE ( gc_actlist ) ;
    MESSAGE "Specify the operation on the summary field";
    WHEN XT_NavDsp( 5 ) ;
    VALID REQUIRED ( TYPE( gc_xsum ) = "C" .AND. ;
                     gc_xoper <> "SUM"             ) .OR. ;
                   ( TYPE( gc_xsum ) = "L" .AND. ;
                     gc_xoper = "CNT" ) .OR.;
                     TYPE( gc_xsum ) $ "NFD";
    ERROR IIF( TYPE( gc_xsum ) = "C", ;
                  "SUM can not be used with character fields", ;
                  "MAX, MIN, or SUM can not be used with logical fields")

RETURN
*-- EOP: XT_DisGets


PROCEDURE XT_IniVars && UDS - CrossTab()
*--------------------------------------------------------------------
* Declare the variables for the CrossTab form as PUBLIC and set
* their values.
*--------------------------------------------------------------------

  PUBLIC gc_xtitl,gc_xtop,gc_xside,gc_xsum,gc_xoper
  gc_xtitl  = SPACE(80)   && Title string for heading of Crosstab
  gc_xtop   = SPACE(80)   && Field for TOP, allow IIF expression
  gc_xside  = SPACE(80)   && Field for SIDE, allow IIF expression
  gc_xsum   = SPACE(80)   && Field to summarize, allow expressions
  gc_xoper  = SPACE(3)    && Operation to perform SUM, CNT, MAX, MIN

*-- Setup for Hand_Pick
  PUBLIC ARRAY aa_lookup[5]
  aa_lookup[1] = ""
  aa_lookup[2] = "FIELDS"
  aa_lookup[3] = "FIELDS"
  aa_lookup[4] = "FIELDS"
  aa_lookup[5] = ""

*-- Setup for Hand_Pick that specifies that the file is already open
  PUBLIC ARRAY params[3]
  params[1] = .T.                      && Lookup file open for this popup
  params[2] = ALIAS()                  && ALIAS of the open lookup file
  params[3] = ALIAS()                  && Current ALIAS that is in USE

RETURN
*-- EOP: XT_IniVars


PROCEDURE XT_TrmVars && UDS - CrossTab()
*-------------------------------------------------------------------
* Trims the trailing blanks from the Crosstab form.
*-------------------------------------------------------------------
  gc_xtitl  = TRIM( gc_xtitl )
  gc_xtop   = TRIM( gc_xtop  )
  gc_xside  = TRIM( gc_xside )
  gc_xsum   = TRIM( gc_xsum  )
  gc_xoper  = TRIM( gc_xoper )
RETURN
*-- EOP: XT_TrmVars


FUNCTION XT_NavDsp  && UDS - CrossTab()
PARAMETER pn_bar
*-------------------------------------------------------------------
* Create the navigation line, and write it as a centered string
*-------------------------------------------------------------------
  PRIVATE lc_navline, ln_navcol

  IF ISBLANK( aa_lookup[ pn_bar ] )
    lc_navline = "Perform:F2  " +;
                 "Save only:Ctrl-End  " +;
                 "Cancel:Esc"
  ELSE
    lc_navline = "Pick operators/fields: Shift-F1  " +;
                 "Perform:F2  " +;
                 "Save only:Ctrl-End  " +;
                 "Cancel:Esc"
  ENDIF
  ln_navcol = Center( lc_navline, 80 )
  lc_win = WINDOW()
  ACTIVATE SCREEN
  SET CURSOR OFF
  IF "43" $ SET("DISPLAY")
    @ 41,0
    @ 41, ln_navcol SAY lc_navline COLOR &messcolor
  ELSE    
    @ 23,0
    @ 23, ln_navcol SAY lc_navline COLOR &messcolor
  ENDIF
  SET CURSOR ON
  ACTIVATE WINDOW &lc_win

RETURN( .T. )
*-- EOF: XT_NavDsp( pn_bar )


PROCEDURE XT_FldPick && UDS - CrossTab()
PARAMETERS pc_fldname
*-------------------------------------------------------------------
* Handle the pick list support for the Crosstab fields.  Actual
* picklist is handled by HAND_PICK in DB4LIB.PRG.
*-------------------------------------------------------------------

*-- Disable the Shift-F1 to prevent a recurrsive call
  ON KEY LABEL Shift-F1
  SET CURSOR OFF

*-- Get the field number for the aa_lookup array
*                        1         2         3         4         5
*               12345678901234567890123456789012345678901234567890
  lc_fldindx = "         GC_XTITL, GC_XTOP,  GC_XSIDE, GC_XSUM,  GC_XOPER"
  ln_fldno   = AT( pc_fldname, lc_fldindx ) / 10

  DO Hand_Pick WITH ln_fldno, .f.   && Call the pick list function

*-- Restore Shift-F1 back
  ON KEY LABEL Shift-F1 DO XT_FldPick WITH VARREAD()
  SET CURSOR ON

RETURN
*-- EOP: XT_FldPick WITH pc_FldName


FUNCTION XT_MakExt  && UDS - CrossTab()
*-------------------------------------------------------------------
* Make the EXECCROS.PRG file if it does not exist.  Add it to the
* current catalog under the Application panel if it's not there.
* If the user is starting CrossTab() from the Dot prompt with no 
* active catalog, XT_MakExt with return a .F. value.
*-------------------------------------------------------------------
  PRIVATE lc_alias, ll_return, ln_catrec

  ll_return = .T.             && Assume the catalog is set up
  lc_alias = ALIAS()                                         

*-- Check to see if EXECCROS.PRG exists, if not create it in current dir
  IF .NOT. FILE( "EXECCROS.PRG" )
    SET ALTERNATE TO execcros.prg
    SET CONSOLE OFF
    SET ALTERNATE ON
    ?
    ?? "PROCEDURE ExecCros"   && Prevent blank line at top of file
    TEXT                      && Dump out the file
*------------------------------------------------------------------
* Execute the Crosstab routine based on the previous QBE with
* the CrossTab() function.
*
* Synopsis:
*
* DO EXECCROS
*
* Description:
*
* EXECCROS will use the current QBE view if the saved value
* in CROSSTAB.MEM (gc_xdbf) matches the currently opened DBF file.
* This will allow EXECCROS to use the correct FILTER is the QBE
* set one.  Otherwise, EXECCROS will close and open data files
* and re-open the DBF file CROSSTAB() used.
*
* Calls:
*
* COM_XTAB.PRG    Calculates the CROSSTAB.DBF based on any filter
*
* Files:
* 
* CROSSTAB.MEM    The memory file with the results from CROSSTAB()
* CROSSTAB.DBF    The resulting file from COM_XTAB.PRG    
*
*------------------------------------------------------------------
  SET TALK OFF
  lc_catalog = SET( "CATALOG" )
  SET CATALOG OFF               
  ll_excl = SET( "EXCLUSIVE" ) = "ON"
  SET EXCLUSIVE ON

*-- Restore the values from the CrossTab() function form
  RESTORE FROM crosstab ADDITIVE

*-- Check to see if the DBF used in CrossTab() is active.  If it is
*-- not, close any open DBF files and open it up.
  gc_filter = ""
  lc_fullp = SET("FULLPATH")
  SET FULLPATH ON
  IF TRIM( DBF() ) <> gc_xdbf
    CLOSE DATABASES
    USE ( gc_xdbf )
  ELSE
    gc_filter = SET( "FILTER" )
    USE ( gc_xdbf )
    IF .NOT. ISBLANK( gc_filter )
      SET FILTER TO &gc_filter
    ENDIF
  ENDIF

*-- Save the VIEW states for COM_XTAB to use
  gc_view   = SET( "VIEW" )

  SET FIELDS OFF                && Prevent problem fo FIELDS not in view
  SET FIELDS TO                 && that were specified in CrossTab()
  SET FULLPATH &lc_fullp.
  lc_safety = SET("SAFETY")     && Set Safety off for overwriting files
  SET SAFETY OFF

*-- Perform the actual crosstab computations
  DO com_xtab WITH "A->"+gc_xtop, "A->"+gc_xside, "A->"+gc_xsum, gc_xoper
  
*-- Display the results of the CrossTab in a BROWSE WINDOW with title
  CLEAR
  DEFINE WINDOW showxtab FROM 5,5 TO 15,76
  @ 3,0 SAY ""
  ?? gc_xtitl AT Center( gc_xtitl, 80 )
  SET DESIGN OFF                    && Prevent transfer to Query design
  BROWSE WINDOW showxtab
  SET DESIGN ON
  @ 3,0
*-- Return the environment back the way we found it
  RELEASE ALL LIKE gc_x*
  RELEASE gn_len, gn_dec, lc_cursor, lc_status, lc_escape, lc_proc
  RELEASE WINDOW showxtab
  SET CATALOG &lc_catalog.
  IF .NOT. ll_excl
    SET EXCLUSIVE OFF
  ENDIF
  SET SAFETY &lc_safety.      
RETURN
*-- EOP: ExecCros.prg
    ENDTEXT                   && End of the output for ExecCros.prg

    SET ALTERNATE OFF         && Close the output file
    SET CONSOLE ON
    SET ALTERNATE TO
  ENDIF

  SELECT ( lc_alias )       && Return the workarea back

RETURN( ll_return )
*-- EOF: XT_MakExt()         



